module Exception(ExT(..),liftEx,raiseEx,catchEx,trapIOEx,
                 transformExUnit,Ex)
where
import Monad

data Exception a b = Raise a | Return b
newtype ExT a m b = ExT (m (Exception a b))

instance Monad m => Monad (ExT a m) where
    return a = liftEx (return a)
    ExT x >>= f = ExT $ do v <- x
                           case v of 
                             Raise e -> return (Raise e)
                             Return w -> do let ExT ret = f w
                                            ret

trapIOEx :: ExT a IO b -> (IOError -> IO a) -> ExT a IO b
trapIOEx (ExT x) f = ExT $ catch (do v <- x
                                     return v)
                                 (\ioerr -> do err <- f ioerr
                                               return (Raise err))


liftEx :: Monad m => m b -> ExT a m b
liftEx m = ExT $ do v <- m
                    return (Return v)

raiseEx :: Monad m => a -> ExT a m b
raiseEx e = ExT (return (Raise e))

catchEx :: Monad m => ExT a m b -> (a -> m b) -> m b
catchEx (ExT try) catch = do v <- try
                             case v of
                               Raise e -> catch e
                               Return w -> return w

transformEx :: (Monad m,Monad n) 
                  => (m (Exception a b) -> n (Exception a b))
                  -> (ExT a m b -> ExT a n b)
transformEx tr (ExT f) = ExT (tr f)

newtype Unit a = Unit a
instance Monad Unit where
   return a = Unit a
   (Unit a) >>= f = f a

transformExUnit :: Monad m => ExT a Unit b -> ExT a m b
transformExUnit = transformEx (\(Unit v) -> return v)

type Ex a b = ExT a Unit b