module Rewrite(calculate) where
import List
import Utils
import Expressions
import Calculations
import SubExpressions
import Substitutions
import Matching
import Names
import Types
import PPrint
import Laws
import Constants
import GList
import FailList
-- import Trace

notrace _ = id

trace=notrace

calculate :: Matcher -> Theor -> Closure -> Calculation
calculate m ceqs (e,bs) 
   = evalContext (e,bs) $ 
          do ceqss <- sequence $ replicate 3 $ scrubTheory $ unletTheor ceqs
             calculate' m ceqss (unlet e,bs)

calculate' :: Matcher -> [Theor] -> Closure -> Name Calculation
calculate' m ceqss (e,bs) = 
   do e' <- etaExp e
      (steps,fails) <- repeatedly (rewrites m ceqss) (e',bs)
      return (etaContract (fst (scrublocs (e',bs))),steps,fails)

rewrites :: Matcher -> [Theor] -> Closure -> Name (FailList Hint Step)
rewrites m ceqss e = do rews <- sequence 
                                 [rewrite m ceqss law se e | 
                                     law <- filter (cviable e) (head ceqss),
                                     se <- filter (flip cviable law . fst) ses]
                        return $ gconcat' rews
       where ses  = filter nontriv (subExps e)
             nontriv ((e,_),_) = not (null (constants e))
             cviable e (_,_,(lhs,_)) = viable m lhs e

repeatedly :: (Closure -> Name (FailList Hint Step)) 
                  -> Closure -> Name ([Step],[Hint])
repeatedly rws (e,bs) 
    = do steps <- rws (e,bs)
         let steps' = gfilter ((/=e).result) steps
         if failnull steps' then return ([],failures steps')
          else do let (n,y) = failhead steps'
                  y' <- etaExp y
                  let y'' = etaContract (fst (scrublocs (y',bs)))
                  (recsteps,fails) <- repeatedly rws (y',bs)
                  return ((n,y''):recsteps,fails)

rewrite :: Matcher -> [Theor] -> CondEq -> SubExp -> 
	       Closure -> Name (FailList Hint Step)
rewrite m ceqss ceq e1 (e,bs) 
    = localContext (e,bs) $ rewritestep m ceqss ceq e1 e


rewritestep :: Matcher -> [Theor] -> CondEq -> SubExp -> Exp 
                 -> Name (FailList Hint Step)
rewritestep m ceqss law@(n,cs,(lhs,rhs)) (e1,loc) e
    =    do ss <- onestepmatch m lhs e1
            let ss' = filter (nontriv lhs) ss
            (if null cs
             then  return $ faillist [] 
                        [(Direct n s, next e loc s rhs) | s <- ss']
             else 
                  do cst <- prove m ceqss cs (bounds e1) ss'
                     return
                        $ gmap 
                           (\(calcs,t) -> 
                              (Indirect n t calcs,next e loc t rhs))
                        $ failedmap (\(calcs,t) -> Indirect n t calcs) cst)
           where next e loc s rhs 
		     = betareduce (replace e loc (appSubst s rhs))
                 nontriv p s = freevars p `subseteq` vardom s

prove :: Matcher -> [Theor] -> [Equation] -> [(VarId,Type)] -> [Subst] 
           -> Name (FailList ([Calculation],Subst) ([Calculation],Subst))
prove m ceqss cs bs ss 
   = pr cs ss
     where pr [] ss = return (faillist [] [([],s) | s <- ss])
           pr (e:es) ss = do cst <- pr es ss  
                             n   <- gmapM f cst
                             return (gconcat n)
            where f (calcs,t) = do (c,d,ts) <- cond m ceqss e bs t
                                   let newcalcs = addcalc c (addcalc d calcs)
                                   return (failmaybe [(newcalcs,t)] 
                                              [(newcalcs,t') | t' <- ts])
                  addcalc calc calcs = if (null (calcsteps calc)) 
                                       then calcs else calc:calcs

cond :: Matcher -> [Theor] -> Equation -> [(VarId,Type)] -> Subst 
                  -> Name (Calculation,Calculation,[Subst])
cond m ceqss (e1,e2) bs s 
          = do ceqss' <- advance ceqss
               c <- calculate' m ceqss' (e1',bs)
               d <- calculate' m ceqss' (e2',bs++tfreevars e2')
               ss <- twostepmatch' m s (conclusion d,bs) 
                                       (conclusion c,bs)
               return (c,d,filter (nontriv d) ss)
                         where e1' = betareduce (appSubst s e1)
                               e2' = betareduce (appSubst s e2)
                               nontriv exp t 
                                 = (freevars (conclusion exp) \\ map fst bs)
                                             `subseteq` vardom t
                               advance [ceqs] = sequence [scrubTheory ceqs]
                               advance (_:ceqss) = return ceqss

etaexp (e,bs) = scrublocs (evalContext (e,bs) (etaExp e),bs)
scrublocs (e,bs) = (evalContext (Con (Number 0) int,bs) (renaming scrub e),bs)

